#include <stdlib.h>
#include <string.h>
#include "types.h"
#include "macros.h"
#include "crc32.h"
#include "iobuf.h"
#include "mtf.h"
#include "rle.h"
#include "sort.h"
#include "bwt.h"


#define RLE_MIN_RUN 4


void          BWTinit(int action, word blockSize);
void          BWTdeInit();
int           BWTcompressFile(int len, FILE *iFile, FILE *oFile);
int           BWTuncompressFile(int len, FILE *iFile, FILE *oFile);
static void   doTransform();
static void   undoTransform();

word          crc, readCRC;
word          bytesIn, bytesOut;
word          cumBytesIn, cumBytesOut;
word          blockSize = K(256), curBlockSize;
word         *stripe;
word         *ptr;
word         *freq;
byte         *block, *block2;

static int    action;
static word   origin;
static word   count[256], cumulative[256];


/**
 **  Allocate memory for compression or decompression
 */
void
BWTinit(int action, word blockSize)
{
    ::action = action;
    ::blockSize = blockSize;
    if (action == BWT_COMPRESS)
    {
        stripe = new word[blockSize + (UNROLL_FACTOR * sizeof(word))];
        ptr = new word[blockSize];
        freq = new word[K(64) + 1];
        block = (byte*)ptr;
    }
    else
    {
        block = new byte[blockSize];
        block2 = new byte[blockSize];
        ptr = new word[blockSize];
    }
}


/**
 **  De-allocate memory.
 */
void
BWTdeInit()
{
    if (action == BWT_COMPRESS)
    {
        delete [] stripe;
        delete [] ptr;
        delete [] freq;
    }
    else
    {
        delete [] block;
        delete [] block2;
        delete [] ptr;
    }
}


/**
 **  Compress a file.  The input stream is fed into a pipeline that
 **  looks like this:
 **
 **      RLE -> BWT -> MTF -> ARITH
 */
int
BWTcompressFile(FILE *iFile, FILE *oFile)
{
    byte *ibuf = new byte[0x2000], *obuf = new byte[0x2000];
    InputBuf iBuf(iFile, ibuf, 0x2000);
    OutputBuf oBuf(NULL, block, blockSize);
    OutputBuf oBuf2(oFile, obuf, 0x2000);

    MTFinit(NULL, &oBuf2);
    CRC32 crc;
    for (;;)
    {
        MTFstartEncoding();
        if (!iBuf.nCharsLeft())
        {
            MTFencodeRandomWord(0xffffffff);
            MTFfinishEncoding();
            break;
        }
        RLE rle(&iBuf, &oBuf, RLE_MIN_RUN);
        rle.encodeBlock();
        curBlockSize = blockSize - oBuf.nCharsLeft();
        crc.addString((byte*)block, curBlockSize);
        doTransform();
        MTFencodeRandomWord(origin);
        MTFencodeBlock();
        MTFfinishEncoding();
        if (curBlockSize < blockSize)
        {
            break;
        }
        oBuf.init(NULL, block, blockSize);
    }
    bytesIn = iBuf.nCharsRead();
    cumBytesIn += bytesIn;
    oBuf2.writeWord(crc.CRC());
    oBuf2.writeWord(bytesIn);
    oBuf2.flush();
    bytesOut += oBuf2.nCharsWritten();
    cumBytesOut += bytesOut;
    MTFdeInit();
    delete ibuf; delete obuf;
    return 0;
}


/**
 **  Decompress a file.  The input stream is fed into a pipeline that
 **  looks like this:
 **
 **      ARITH -> MTF -> BWT -> RLE
 */
int
BWTuncompressFile(FILE *iFile, FILE *oFile)
{
    byte *ibuf = new byte[0x2000], *obuf = new byte[0x2000];
    InputBuf iBuf(iFile, ibuf, 0x2000);
    OutputBuf oBuf(NULL, block, blockSize);
    OutputBuf oBuf2(oFile, obuf, 0x2000);

    MTFinit(&iBuf, &oBuf);
    CRC32 crc;
    for (;;)
    {
        MTFstartDecoding();
        origin = MTFdecodeRandomWord();
        if (origin == 0xffffffff)
        {
            break;
        }
        MTFdecodeBlock();
        undoTransform();
        crc.addString(block2, curBlockSize);
        if (oFile)
        {
            InputBuf iBuf2(NULL, block2, curBlockSize);
            RLE rle(&iBuf2, &oBuf2, RLE_MIN_RUN);
            rle.decodeBlock();
        }
        if (curBlockSize < blockSize)
        {
            break;
        }
    }
    oBuf2.flush();
    ::crc = crc.CRC();
    iBuf.readWord(&readCRC);
    MTFdeInit();
    delete ibuf; delete obuf;
    return 0;
}


/**
 **  Doing the transform consists of sorting the strings in order to
 **  determine L[], and determining the origin.  The origin is the index
 **  into L[] where we can find the first character in the block.  When
 **  we are undoing the transform, we will start at this point in order
 **  to reconstruct the block.
 */
void
doTransform()
{
    word i;

    doSort();
    for (i = 0; i < curBlockSize; i++)
    {
        if (ptr[i] == 1)
        {
            origin = i;
            break;
        }
    }
}


/**
 **  The BWT relies on the fact that for any multiple occurances of a
 **  character in L[], the corresponding strings will occur *in the same
 **  order* (though not necessarily in the same places) in F[].  This means
 **  that we can determine, for each character in L[], what its *predecessor*
 **  string is.  We create ptr[], such that ptr[i] is the successor of string
 **  i.  Given origin, L[], and ptr[], we can easily re-create the original
 **  block.
 */
void
undoTransform()
{
    word i, j, c, sum;

    // Determine count[].  count[i] = #/occurances of character i.
    memset(count, 0, 256 * sizeof(word));
    for (i = 0; i < curBlockSize; i++)
    {
        count[block[i]]++;
    }
    // Determine cumulative[].
    // cumulative[i] = count[0] + count[1] + ... + count[i - 1]
    for (sum = 0, i = 0; i < 256; i++)
    {
        cumulative[i] = sum;
        sum += count[i];
        count[i] = 0;
    }
    // Determine ptr[].
    for (i = 0; i < curBlockSize; i++)
    {
        c = block[i];
        ptr[cumulative[c] + count[c]] = i;
        count[c]++;
    }
    // Now we can re-create the original block.
    j = origin;
    for (i = 0; i < curBlockSize; i++)
    {
        block2[i] = block[j];
        j = ptr[j];
    }
}
