#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "types.h"
#include "macros.h"
#include "bwt.h"
#include "sort.h"


static void         doStripe();
static void         quickSort(word l, word h);
static int          boundedCompare(word *_i1, word *_i2);
static inline int   inlineBoundedCompare(word i1, word i2);


/**
 **  A clever trick is used to speed the sort operation.  We first radix sort
 **  on the first two bytes of each string.  Then we sort the resulting
 **  buckets in increasing order of their size.  In the interest of speeding
 **  up the sorting of future (larger) buckets, we "remember" how the strings
 **  in the current bucket rank against each other.  When comparing strings in
 **  a future bucket, we may come to a point where we are comparing strings
 **  which have already been compared, and in this case we will instantly know
 **  the result.  Without this optimization, a much larger penalty is incurred
 **  when sorting large blocks.  There is no guarantee that this will help
 **  much of course -- we could still end up with *only* large buckets
 **  (size > 64K), and in this case we would have no opportunity to use this
 **  optimization.  However, the optimization is effective on the vast
 **  majority of real-world data.
 */
void doSort()
{
    word i, j, k, phase, minSize, maxSize, l, h, f;

    doStripe();
    if (curBlockSize <= 64)
    {
        for (i = 0; i < curBlockSize; i++)
        {
            ptr[i] = i;
        }
        qsort(ptr,
              curBlockSize,
              sizeof(word),
              (int (*)(const void*, const void*))boundedCompare);
        return;
    }
    memset(freq, 0, (K(64) + 1) * sizeof(word));
    for (i = 0; i < curBlockSize; i++)
    {
        freq[stripe[i] >> 16]++;
    }
    for (i = 1; i <= K(64); i++)
    {
        freq[i] += freq[i - 1];
    }
    for (i = 0; i < curBlockSize; i++)
    {
        j = stripe[i] >> 16;
        ptr[--freq[j]] = i;
    }

    for (phase = 0; phase <= 5; phase++)
    {
        switch (phase)
        {
        case 0: minSize = 2;          maxSize = 16;         break;
        case 1: minSize = 17;         maxSize = 256;        break;
        case 2: minSize = 257;        maxSize = K(4);       break;
        case 3: minSize = K(4) + 1;   maxSize = K(32);      break;
        case 4: minSize = K(32) + 1;  maxSize = K(64);      break;
        case 5: minSize = K(64) + 1;  maxSize = K(K(256));  break;
        }
        for (i = 0; i < K(64); i++)
        {
            l = freq[i];
            f = freq[i + 1] - l;
            if (f >= minSize && f <= maxSize)
            {
                h = l + f - 1;
                quickSort(l, h);
                if (f <= K(64))
                {
                    for (k = 0, j = l; j <= h; j++)
                    {
                        stripe[ptr[j]] &= 0xffff0000;
                        stripe[ptr[j]] |= k++;
                        if (ptr[j] < (UNROLL_FACTOR * sizeof(word)))
                        {
                            stripe[curBlockSize + ptr[j]] = stripe[ptr[j]];
                        }
                    }
                }
            }
        }
    }
}


void doStripe()
{
    word  i;
    byte  c;
    int   i1, i2, i3;

    bzero(stripe, curBlockSize * sizeof(word));
    for (i = 0; i < curBlockSize; i++)
    {
        c = block[i];
        i1 = i - 1; if (i1 < 0) i1 += curBlockSize;
        i2 = i - 2; if (i2 < 0) i2 += curBlockSize;
        i3 = i - 3; if (i3 < 0) i3 += curBlockSize;
        stripe[i] |= c << 24;
        stripe[i1] |= c << 16;
        stripe[i2] |= c << 8;
        stripe[i3] |= c;
    }
    for (i = curBlockSize;
         i < curBlockSize + (UNROLL_FACTOR * sizeof(word));
         i++)
    {
        stripe[i] = stripe[i - curBlockSize];
    }
}


inline
int
inlineBoundedCompare(word i1, word i2)
{
    word nComps, maxComps;
    word w1, w2, *wp1, *wp2, *wpLimit = stripe + curBlockSize;

    if (i1 == i2)
    {
        return 0;
    }

    maxComps = curBlockSize / sizeof(word);
    if (curBlockSize & (sizeof(word) - 1))
    {
        maxComps++;
    }

    nComps = 0;
    wp1 = stripe + i1;
    wp2 = stripe + i2;
    for (nComps = 0;;)
    {
        w1 = *wp1;
        w2 = *wp2;
        if (w1 < w2) return -1;
        else if (w1 > w2) return 1;
        wp1++; wp2++;

        w1 = *wp1;
        w2 = *wp2;
        if (w1 < w2) return -1;
        else if (w1 > w2) return 1;
        wp1++; wp2++;

        w1 = *wp1;
        w2 = *wp2;
        if (w1 < w2) return -1;
        else if (w1 > w2) return 1;
        wp1++; wp2++;

        w1 = *wp1;
        w2 = *wp2;
        if (w1 < w2) return -1;
        else if (w1 > w2) return 1;
        wp1++; wp2++;

        w1 = *wp1;
        w2 = *wp2;
        if (w1 < w2) return -1;
        else if (w1 > w2) return 1;
        wp1++; wp2++;

        w1 = *wp1;
        w2 = *wp2;
        if (w1 < w2) return -1;
        else if (w1 > w2) return 1;
        wp1++; wp2++;

        w1 = *wp1;
        w2 = *wp2;
        if (w1 < w2) return -1;
        else if (w1 > w2) return 1;
        wp1++; wp2++;

        w1 = *wp1;
        w2 = *wp2;
        if (w1 < w2) return -1;
        else if (w1 > w2) return 1;
        wp1++; wp2++;

        w1 = *wp1;
        w2 = *wp2;
        if (w1 < w2) return -1;
        else if (w1 > w2) return 1;
        wp1++; wp2++;

        w1 = *wp1;
        w2 = *wp2;
        if (w1 < w2) return -1;
        else if (w1 > w2) return 1;
        wp1++; wp2++;

        w1 = *wp1;
        w2 = *wp2;
        if (w1 < w2) return -1;
        else if (w1 > w2) return 1;
        wp1++; wp2++;

        w1 = *wp1;
        w2 = *wp2;
        if (w1 < w2) return -1;
        else if (w1 > w2) return 1;
        wp1++; wp2++;

        w1 = *wp1;
        w2 = *wp2;
        if (w1 < w2) return -1;
        else if (w1 > w2) return 1;
        wp1++; wp2++;

        w1 = *wp1;
        w2 = *wp2;
        if (w1 < w2) return -1;
        else if (w1 > w2) return 1;
        wp1++; wp2++;

        w1 = *wp1;
        w2 = *wp2;
        if (w1 < w2) return -1;
        else if (w1 > w2) return 1;
        wp1++; wp2++;

        w1 = *wp1;
        w2 = *wp2;
        if (w1 < w2) return -1;
        else if (w1 > w2) return 1;
        wp1++; wp2++;

        nComps += 16;
        if (nComps >= maxComps)
        {
            return 0;
        }

        if (wp1 >= wpLimit)
        {
            wp1 -= curBlockSize;
        }
        if (wp2 >= wpLimit)
        {
            wp2 -= curBlockSize;
        }
    }

    return 0;
};


/**
 **  Quicksort, with median-of-three paritioning.
 */
#define SWAP(x, y)  { temp = ptr[x] ; ptr[x] = ptr[y] ; ptr[y] = temp; }

void quickSort(word l, word h)
{
    word          i, j, k, mid, p;
    register word temp;
    word          stack[128];
    word         *sp = stack;

    for (;;)
    {
        if (h - l >= 8)
        {
            mid = (l + h) >> 1;
            if (inlineBoundedCompare(ptr[l], ptr[mid]) > 0)
            {
                SWAP(l, mid);
            }
            if (inlineBoundedCompare(ptr[mid], ptr[h]) > 0)
            {
                SWAP(mid, h);
            }
            if (inlineBoundedCompare(ptr[l], ptr[mid]) > 0)
            {
                SWAP(l, mid);
            }
            SWAP(mid, h - 1);
            p = ptr[h - 1];
            i = l; j = h - 1;
            for (;;)
            {
                for (;;)
                {
                    i++;
                    if (inlineBoundedCompare(ptr[i], p) >= 0)
                    {
                        break;
                    }
                }
                for (;;)
                {
                    j--;
                    if (inlineBoundedCompare(ptr[j], p) <= 0)
                    {
                        break;
                    }
                }
                if (i >= j)
                {
                    break;
                }
                SWAP(i, j);
            }
            SWAP(i, h - 1);
            if (i - l > h - i)
            {
                *sp++ = l;
                *sp++ = i - 1;
                l = i + 1;
            }
            else
            {
                *sp++ = i + 1;
                *sp++ = h;
                h = i - 1;
            }
        }
        else
        {
            i = l;
            while (++i <= h)
            {
                j = ptr[i];
                k = i;
                for (;;)
                {
                    if (inlineBoundedCompare(j, ptr[k - 1]) >= 0)
                    {
                        break;
                    }
                    ptr[k] = ptr[k - 1];
                    if (--k == l)
                    {
                        break;
                    }
                }
                ptr[k] = j;
            }
            if (sp == stack)
            {
                return;
            }
            h = *(--sp); l = *(--sp);
        }
    }
}


int
boundedCompare(word *_i1, word *_i2)
{ 
    word  i1 = *_i1,
          i2 = *_i2,
          nComps,
          maxComps,
          w1, w2;

    if (i1 == i2)
    {
        return 0;
    }

    maxComps = curBlockSize / sizeof(word);
    if (curBlockSize & (sizeof(word) - 1))
    {
        maxComps++;
    }

    for (nComps = 0; nComps < maxComps; nComps++)
    {
        w1 = stripe[i1];
        w2 = stripe[i2];
        if (w1 < w2) return -1;
        else if (w1 > w2) return 1;
        i1 += sizeof(word); if (i1 >= curBlockSize) i1 -= curBlockSize;
        i2 += sizeof(word); if (i2 >= curBlockSize) i2 -= curBlockSize;
    }

    return 0;
};
