#include <getopt.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
#include "types.h"
#include "macros.h"
#include "crc32.h"
#include "bwt.h"
#include "iobuf.h"
#include "mtf.h"
#include "rle.h"


#define TZIP_COMPRESS       0
#define TZIP_UNCOMPRESS     1
#define TZIP_TEST           2


const char     *version = "tzip v1.12 (Aug 1 1998) by Adam McKee";


static int      compressFile(FILE *iFile, FILE *oFile);
static int      uncompressFile(FILE *iFile, FILE *oFile);
static void     warn(const char *fmt, ...);
static void     pwarn(const char *fmt, ...);
static void     errQuit(const char *fmt, ...);
static void     help();


static char     iPath[256], oPath[256];
static int      action;
static int      optStdout;
static int      optForce;
static int      optPreserve;
static int      optQuiet;
static int      optVerbose;
static Boolean  headingWritten;
static char     myName[17];


int main(int argc, char **argv)
{
    int           i, res = 0;
    struct stat   statBuf;
    FILE         *iFile, *oFile;

    i = strlen(argv[0]);
    if (i >= 4 && !strcmp(argv[0] + i - 4, "tzip"))
    {
        strcpy(myName, "tzip");
        action = TZIP_COMPRESS;
    }
    else
    {
        strcpy(myName, "tunzip");
        action = TZIP_UNCOMPRESS;
    }

    if ((argc == 2) && (strcmp(argv[1], "--help") == 0))
    {
        help();
    }

    while ((i = getopt(argc, argv, "Vbcdfhipqtv123456789")) != EOF)
    {
        switch (i)
        {
        case 'V': printf("%s\n", version); exit(0);
        case 'b': blockSize = K(atoi(argv[optind++])); break;
        case 'c': optStdout = 1; optPreserve = 1; break;
        case 'd': action = TZIP_UNCOMPRESS; break;
        case 'f': optForce = 1; break;
        case 'h': help(); break;
        case 'i': incVal[atoi(argv[optind])] = atoi(argv[optind+1]); optind += 2; break;
        case 'p': optPreserve = 1; break;
        case 'q': optQuiet = 1; break;
        case 't': action = TZIP_TEST;
        case 'v': optVerbose = 1; break;
        case '1': blockSize = K(64); break;
        case '2': blockSize = K(128); break;
        case '3': blockSize = K(256); break;
        case '4': blockSize = K(384); break;
        case '5': blockSize = K(512); break;
        case '6': blockSize = K(640); break;
        case '7': blockSize = K(768); break;
        case '8': blockSize = K(896); break;
        case '9': blockSize = K(1024);
        }
    }

    if (optind == argc)
    {
        optStdout = 1;
    }

    if ((action == TZIP_COMPRESS) && optStdout && isatty(STDOUT_FILENO))
    {
        errQuit("\
compressed data not written to terminal.\n\
For help, type: %s -h", myName);
    }

    if (optind < argc)
    {
        for (i = optind; i < argc; i++)
        {
            if (action == TZIP_COMPRESS)
            {
                if ((strlen(argv[i]) >= 3) &&
                    (strcasecmp(argv[i] + strlen(argv[i]) - 3, ".tz") == 0))
                {
                    warn("`%s' already has .tz extension -- skipping",
                         argv[i]);
                    continue;
                }
            }
            else
            {
                if ((strlen(argv[i]) < 3) ||
                    (strcasecmp(argv[i] + strlen(argv[i]) - 3, ".tz") != 0))
                {
                    warn("`%s' does not have .tz extension -- skipping",
                         argv[i]);
                    continue;
                }
            }

            strcpy(iPath, argv[i]);

            if (lstat(iPath, &statBuf) < 0)
            {
                pwarn("couldn't stat `%s'", iPath);
                continue;
            }
            if (S_ISDIR(statBuf.st_mode))
            {
                warn("`%s' is a directory -- skipping", iPath);
                continue;
            }
            if (S_ISLNK(statBuf.st_mode) && !optForce)
            {
                warn("`%s' is a symbolic link -- skipping", iPath);
                continue;
            }

            iFile = fopen(iPath, "rb");
            if (!iFile)
            {
                pwarn("couldn't open `%s'", iPath);
                continue;
            }

            if (optStdout)
            {
                strcpy(oPath, "<stdout>");
                oFile = stdout;
            }
            else
            {
                if (action == TZIP_COMPRESS)
                {
                    sprintf(oPath, "%s.tz", argv[i]);
                }
                else if (action == TZIP_UNCOMPRESS)
                {
                    strcpy(oPath, argv[i]);
                    oPath[strlen(oPath) - 3] = '\0';
                }
                else
                {
                    oPath[0] = '\0';
                }

                if (oPath[0] != '\0')
                {
                    if (!stat(oPath, &statBuf) && !optForce)
                    {
                        fclose(iFile);
                        warn("`%s' already exists -- skipping", oPath);
                        continue;
                    }

                    oFile = fopen(oPath, "wb");
                    if (!oFile)
                    {
                        fclose(iFile);
                        pwarn("couldn't open `%s'", oPath);
                        continue;
                    }
                }
                else
                {
                    oFile = NULL;
                }
            }

            if (action == TZIP_COMPRESS)
            {
                res = compressFile(iFile, oFile);
            }
            else
            {
                res = uncompressFile(iFile, oFile);
            }

            fclose(iFile);

            if ((optStdout) && (action == TZIP_COMPRESS))
            {
                exit(0);
            }

            if (!optStdout && (oPath[0] != '\0'))
            {
                fclose(oFile);
            }

            if (res)
            {
                if (action == TZIP_UNCOMPRESS)
                {
                    unlink(oPath);
                }
            }

            if (!optPreserve && (action != TZIP_TEST))
            {
                if (unlink(iPath))
                {
                    pwarn("couldn't unlink `%s'", iPath);
                    continue;
                }
            }
        }
    }
    else
    {
        strcpy(iPath, "<stdin>");
        strcpy(oPath, "<stdout>");
        if (action == TZIP_COMPRESS)
        {
            res = compressFile(stdin, stdout);
        }
        else if (action == TZIP_UNCOMPRESS)
        {
            res = uncompressFile(stdin, stdout);
        }
        else
        {
            res = uncompressFile(stdin, NULL);
        }
    }
    if (optVerbose && action == TZIP_COMPRESS && headingWritten)
    {
        iPath[0] = '\0';
        printf("------------------------------------------------------------------------------\n");
        printf("%-36s%-12u%-12u%-11.2f%.2f\n",
               iPath,
               cumBytesIn,
               cumBytesOut,
               (float)cumBytesOut / cumBytesIn * 100,
               (float)cumBytesOut / cumBytesIn * 8);
    }
    exit(0);
}


int
compressFile(FILE *iFile, FILE *oFile)
{
    int   len = 7;
    byte  buf[7];
    char  path[256];

    OutputBuf oBuf(oFile, buf, len);
    oBuf.putChar('T');
    oBuf.putChar('Z');
    oBuf.putChar(0);
    oBuf.writeWord(blockSize);
    oBuf.flush();
    bytesOut = 7;
    BWTinit(BWT_COMPRESS, blockSize);
    BWTcompressFile(iFile, oFile);
    BWTdeInit();
    if (strlen(iPath) > 34)
    {
        sprintf(path, "...%s", iPath + strlen(iPath) - 34 + 3);
    }
    else
    {
        strcpy(path, iPath);
        if (optVerbose)
        {
            if (!headingWritten)
            {
                printf("\
Pathname                            BytesIn     BytesOut    Perc.    Bits/Byte\n");
                printf("\
------------------------------------------------------------------------------\n");
                headingWritten = TRUE;
            }
            printf("%-36s%-12u%-12u%-11.2f%.2f\n",
                   path,
                   bytesIn,
                   bytesOut,
                   (float)bytesOut / bytesIn * 100,
                   (float)bytesOut / bytesIn * 8);
        }
    }
    return 0;
}


int
uncompressFile(FILE *iFile, FILE *oFile)
{
    int   len = 7;
    byte  buf[7];

    InputBuf iBuf(iFile, buf, len);
    if (iBuf.getChar() != 'T' || iBuf.getChar() != 'Z')
    {
        warn("`%s' is not in tzip format -- skipping", iPath);
        return 1;
    }
    if (iBuf.getChar() != 0)
    {
        warn("`%s' is in an unsupported format -- skipping", iPath);
        return 1;
    }
    iBuf.readWord(&blockSize);
    BWTinit(BWT_UNCOMPRESS, blockSize);
    BWTuncompressFile(iFile, oFile);
    BWTdeInit();
    if (crc != readCRC) {
        warn("CRC failed for `%s'", iPath);
        return 1;
    }
    return 0;
}


void
warn(const char *fmt, ...)
{
    char      buf[256];
    va_list   ap;

    if (optQuiet)
    {
        return;
    }
    va_start(ap, fmt);
    vsprintf(buf, fmt, ap);
    fprintf(stderr, "%s: ", myName);
    fprintf(stderr, "%s\n", buf);
    va_end(ap);
}


void
pwarn(const char *fmt, ...)
{
    char     buf[256], buf1[256];
    va_list  ap;

    if (optQuiet)
    {
        return;
    }
    va_start(ap, fmt);
    vsprintf(buf, fmt, ap);
    sprintf(buf1, "%s: %s", myName, buf);
    perror(buf1);
    va_end(ap);
}


void
errQuit(const char *fmt, ...)
{
    char      buf[256];
    va_list   ap;

    va_start(ap, fmt);
    vsprintf(buf, fmt, ap);
    fprintf(stderr, "%s: ", myName);
    fprintf(stderr, "%s\n", buf);
    va_end(ap);
    exit(1);
}


void
help()
{
    printf("\
%s\n\
usage: %s [-Vbcdhpqtv123456789] [name ...]\n\
        -V   display version number and exit\n\
        -b   set block size in kilobytes\n\
        -c   write to standard output (implies -p)\n\
        -d   decompress\n\
        -f   overwrite existing files and compress links\n\
        -h   show this help\n\
        -p   preserve (don't unlink) input files\n\
        -q   suppress warnings\n\
        -t   test compressed file integrity\n\
        -v   be verbose\n\
        -1   use 64K block size (compress least, use least memory)\n\
        -2   use 128K block size\n\
        -3   use 256K block size (default)\n\
        -4   use 384K block size\n\
        -5   use 512K block size\n\
        -6   use 640K block size\n\
        -7   use 768K block size\n\
        -8   use 896K block size\n\
        -9   use 1MB block size (compress most, use most memory)\n\
[name ...]   files to (de)compress (stdin is read if none are given)\n", version, myName);
    exit(1);
}
